
/*
	OneStroke -- onestroke.c
	Copyright 2007 Michael Levin <risujin@risujin.org>

	NOTE: This project has been succeeded by CellWriter, a grid-entry
	      handwriting input panel. Please see:
	        http://risujin.org/cellwriter
*/

/*
    This file is part of OneStroke.

    OneStroke is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    OneStroke is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OneStroke; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/* Standard library headers */
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <values.h>

/* Debug headers */
#ifdef _DEBUG
#include <stdarg.h>
#include <stdio.h>
#endif

/* OneStroke header */
#include "onestroke.h"

/*
	Symbol Set
*/

static osSymbol *symset_first,*symset_last;

osSymbol* osSymsetFirst(void) {
	return symset_first;
}

osSymbol* osSymsetLast(void) {
	return symset_last;
}

void osSymsetAdd(osSymbol* symbol) {
	if(osSymValid(symbol)) {
		if(!symset_first) {
			assert(!symset_last);
			symset_first = symbol;
		} else {
			assert(symset_last);
			symset_last->next = symbol;
			symbol->prev = symset_last;
		}
		symset_last = symbol;
		assert(symbol->next==0);
		os_modified = 1;
	}
}

void osSymsetFree(void) {
	while(symset_first) osSymFree(symset_first);
	assert(symset_last==0);
	os_modified = 1;
}


/*
	Current Stroke
*/

#define STROKE_INC 64

osSymbol *os_cur_stroke;
float os_match_cx=0.,os_match_cy=0.;
static int cur_stroke_size = 0;
int os_mode = OS_TRAP_COMP;

void osCurDraw(osPoint pt) {

	/* Resize stroke array if necessary */
	if(os_cur_stroke->length>=cur_stroke_size) {
		cur_stroke_size += STROKE_INC;
		debug("Resizing stroke array to %d entries.",cur_stroke_size);
		os_cur_stroke->points = realloc(os_cur_stroke->points,
			cur_stroke_size*sizeof(osPoint));
	}

	/* Update distance */
	if(os_cur_stroke->length>0) {
		double x = os_cur_stroke->points[os_cur_stroke->length-1].x-pt.x,
			y = os_cur_stroke->points[os_cur_stroke->length-1].y-pt.y;
		os_cur_stroke->distance += sqrt(x*x+y*y);
	}

	os_cur_stroke->points[os_cur_stroke->length++] = pt;
}

void osCurReset(void) {
	os_cur_stroke->length = 0;
	os_cur_stroke->distance = 0.;
}

/* Absolute value of the smallest difference between two angles */
static float angleAbsDif(float a,float b) {
	float dif=a-b;
	if(dif>=M_PI) dif = 2.f*M_PI-dif;
	assert(dif<M_PI);
	if(dif<-M_PI) dif += 2.f*M_PI;
	assert(dif>=-M_PI);
	return dif;
}

static float angleSum(float a,float b) {
	float sum=a+b;
	if(sum<0.f) sum += 2.f*M_PI;
	assert(sum>=0.f);
	if(sum>=2.f*M_PI) sum -= 2.f*M_PI;
	assert(sum<2.f*M_PI);
	return sum;
}

static osSymbol* align(osSymbol *symbol,osSymbol *to,float *pangle) {
	osSymbol *aligned = osSymNew();
	float angle = angleSum(to->direction,-symbol->direction),
		sin_a = sinf(-angle),cos_a = cosf(-angle);
	int i;

	/* Copy symbol */
	aligned->distance = symbol->distance;
	aligned->key = symbol->key;
	aligned->flags = symbol->flags;
	aligned->length = symbol->length;
	aligned->points = malloc(aligned->length*sizeof(osPoint));
	aligned->segments = malloc((aligned->length-1)*sizeof(osSegment));

	/* Copy and rotate symbol points/segments */
	debug("Rotating symbol (%.3f) to (%.3f) by %.3f rad.",
		symbol->direction,to->direction,angle);
	for(i=0;i<symbol->length;i++) {

		/* Rotate point */
		aligned->points[i].x = (symbol->points[i].x-OS_SCALE/2)*cos_a-
			(symbol->points[i].y-OS_SCALE/2)*sin_a+OS_SCALE/2;
		aligned->points[i].y = (symbol->points[i].x-OS_SCALE/2)*sin_a+
			(symbol->points[i].y-OS_SCALE/2)*cos_a+OS_SCALE/2;

		/* Rotate segment */
		if(i<symbol->length-1) {
			aligned->segments[i].length = symbol->segments[i].length;
			aligned->segments[i].angle =
				angleSum(symbol->segments[i].angle,angle);
		}
	}

	*pangle = angle;
	return aligned;
}

osSymbol* osCurMatch(void) {
	static osSymbol *stored=0;
	float best_dist = MAXFLOAT;
	osSymbol *sym,*best,*matching=0;

	/* Process current stroke */
	osSymProcess(os_cur_stroke,&os_match_cx,&os_match_cy);

	/* Reject one point strokes */
	if(os_cur_stroke->length<=1) return 0;

	/* Compare the current stroke to every stored stroke */
	sym = osSymsetFirst();
	while(sym) {
		if(sym->length) {
			float dist,angle=1.f;

			/* Align stored stroke to current stroke */
			if(os_mode&OS_ROTATE) {
				matching = align(sym,os_cur_stroke,&angle);
				angle = angleAbsDif(angle,0)+1.f;
			} else matching = sym;

			/* Run distance algorithm */
			switch(os_mode&OS_MODE_MASK) {
				case OS_TRAP_DIST:
					dist = osSymTrapezoid(os_cur_stroke,matching);
					break;
				case OS_TRAP_ANGLE:
					dist = osSymTrapAngle(os_cur_stroke,matching);
					break;
				case OS_TRAP_COMP:
					dist = osSymTrapComp(os_cur_stroke,matching);
					break;
				default:
					osSymFree(matching);
					debug("Invalid matching mode %d.",os_mode&OS_MODE_MASK);
					return 0;
			}
			if((os_mode&OS_ROTATE)&&(os_mode&OS_FAVOR)) dist *= angle;
			if(dist<best_dist) {
				osSymFree(stored);
				best_dist = dist;
				best = matching;
				if(matching!=sym) stored = matching;
				else stored = 0;
			} else if(matching!=sym) osSymFree(matching);
			#ifdef _DEBUG
			{
				char *str = osSymKeyString(sym);
				debug("Distance to symbol '%s' is %.4f.",str,dist);
				free(str);
			}
			#endif
		}
		sym = osSymsetNext(sym);
	}
	return best;
}


/*
	Symbols
*/

#define OS_THRESHOLD 0.5

static void simplify(osSymbol *symbol) {
	int i;
	/*#ifdef _DEBUG
	int old = symbol->length;
	#endif*/

	/* Find node with the least distance to the line of its neighbors */
	for(i=1;i<symbol->length-1;i++) {
		double lx,ly,lm,wx,wy,dist;
		lx = symbol->points[i-1].x-symbol->points[i+1].x;
		ly = symbol->points[i-1].y-symbol->points[i+1].y;
		lm = sqrt(lx*lx+ly*ly);
		lx /= lm;
		ly /= lm;
		wx = symbol->points[i-1].x-symbol->points[i].x;
		wy = symbol->points[i-1].y-symbol->points[i].y;
		dist = lx*wy-ly*wx;
		if(dist<0.) dist = -dist;

		/* Crush node is below threshold */
		if(dist<OS_THRESHOLD) {
			symbol->length--;
			memcpy(symbol->points+i,symbol->points+i+1,
				(symbol->length-i)*sizeof(osPoint));
			i--;
		}
	}

	/*debug("Crushed %d nodes (%d remaining).",
		old-symbol->length,symbol->length);*/
}

static void weightCenter(osSymbol *sym,float *pcx,float *pcy) {
	float cx=0.,cy=0.;
	int i;

	/* Locate weighted center */
	for(i=0;i<sym->length-1;i++) {
		cx += (sym->points[i].x+sym->points[i+1].x)*sym->segments[i].length/2;
		cy += (sym->points[i].y+sym->points[i+1].y)*sym->segments[i].length/2;
	}
	cx /= sym->distance;
	cy /= sym->distance;

	/* Shift all the points if off-center */
	if(cx<OS_SCALE/2-1||cx>OS_SCALE/2+1||cy<OS_SCALE/2-1||cy>OS_SCALE/2+1) {
		/*debug("Recentering symbol by %.1f, %.1f.",
			OS_SCALE/2-cx,OS_SCALE/2-cy);*/
		for(i=0;i<sym->length;i++) {
			float nx,ny;
			nx = (float)sym->points[i].x+OS_SCALE/2-cx;
			ny = (float)sym->points[i].y+OS_SCALE/2-cy;
			if(nx<0.f) nx = 0.f;
			else if(nx>OS_SCALE) nx = OS_SCALE;
			if(ny<0.f) ny = 0.f;
			else if(ny>OS_SCALE) ny = OS_SCALE;
			sym->points[i].x = (unsigned char)nx;
			sym->points[i].y = (unsigned char)ny;
		}
		if(pcx&&pcy) {
			*pcx = cx;
			*pcy = cy;
		}
	} else {
		/*debug("Symbol already weight centered.");*/
		if(pcx&&pcy) {
			*pcx = 0;
			*pcy = 0;
		}
	}
}

void osSymProcess(osSymbol *symbol,float *cx,float *cy) {
	int i;

	/* Simplify the symbol */
	simplify(symbol);

	/* Create and compile segments array */
	if(symbol->segments) free(symbol->segments);
	symbol->segments = malloc((symbol->length-1)*sizeof(osSegment));
	for(i=0;i<symbol->length-1;i++) {
		float dx = symbol->points[i+1].x-symbol->points[i].x,
			dy = symbol->points[i].y-symbol->points[i+1].y;
		symbol->segments[i].angle = atan2f(dy,dx);
		if(symbol->segments[i].angle<0.f) symbol->segments[i].angle += M_PI*2.f;
		symbol->segments[i].length = sqrtf(dx*dx+dy*dy);
	}

	/* Weight-center the symbol */
	weightCenter(symbol,cx,cy);

	/* Symbol direction */
	symbol->direction = atan2f(symbol->points[0].y-OS_SCALE/2,
		OS_SCALE/2-symbol->points[0].x);
	/*debug("Symbol direction is %f.",symbol->direction);*/
}

osSymbol* osSymNew(void) {
	osSymbol *symbol = malloc(sizeof(osSymbol));
	osSymInit(symbol);
	return symbol;
}

void osSymInit(osSymbol* sym) {
	if(osSymValid(sym)) memset(sym,0,sizeof(osSymbol));
}

void osSymFree(osSymbol* sym) {
	if(osSymValid(sym)) {
		if(sym==symset_first) {
			symset_first = sym->next;
			os_modified = 1;
		}
		if(sym==symset_last) {
			symset_last = sym->prev;
			os_modified = 1;
		}
		if(sym->next) {
			sym->next->prev = sym->prev;
			os_modified = 1;
		}
		if(sym->prev) {
			sym->prev->next = sym->next;
			os_modified = 1;
		}
		if(sym->points) free(sym->points);
		free(sym);
	}
}

void osSymSetFree(osSymbol *dest,osSymbol *src) {
	if(osSymValid(dest)&&osSymValid(src)) {
		dest->points = src->points;
		dest->length = src->length;
		dest->distance = src->distance;
		dest->key = src->key;
		dest->flags = src->flags;
		free(src);
		os_modified = 1;
	}
}

void osSymKeyChange(osSymbol *symbol,int key,short flags) {
	if(osSymValid(symbol)&&(symbol->key!=key||symbol->flags!=flags)) {
		symbol->key = key;
		symbol->flags = flags;
		os_modified = 1;
	}
}

static void createCompat(osSymbol *a,osSymbol *b,osSymbol *ac,osSymbol *bc) {
	int i,j,k,running=1;
	float len,dist_a=0.f,dist_b=0.f;

	/* Allocate memory */
	len = a->length+b->length;
	ac->distance = a->distance;
	bc->distance = b->distance;
	ac->points = malloc(len*sizeof(osPoint));
	bc->points = malloc(len*sizeof(osPoint));
	ac->segments = malloc((len-1)*sizeof(osSegment));
	bc->segments = malloc((len-1)*sizeof(osSegment));

	/* First points */
	ac->points[0] = a->points[0];
	bc->points[0] = b->points[0];
	for(i=0,j=0,k=1;running;k++) {
		float da = a->segments[i].length-dist_a,
			db = b->segments[j].length-dist_b,
			pa = da/a->distance,pb = db/b->distance;

		/* Set angles */
		if(i<a->length-1&&j<b->length-1) {
			ac->segments[k-1].angle = a->segments[i].angle;
			bc->segments[k-1].angle = b->segments[j].angle;
			assert(ac->segments[k-1].angle>-2.f*M_PI);
			assert(ac->segments[k-1].angle<2.f*M_PI);
			assert(bc->segments[k-1].angle>-2.f*M_PI);
			assert(bc->segments[k-1].angle<2.f*M_PI);
		} else running = 0;

		/* Advance to the closest point and interpolate the other point */
		if(pa<pb) {
			dist_b += pa*b->distance;
			dist_a = 0.f;
			ac->segments[k-1].length = da;
			bc->segments[k-1].length = pa*b->distance;
			ac->points[k] = a->points[++i];
			bc->points[k].x = (b->points[j+1].x*dist_b+b->points[j].x*
				(b->segments[j].length-dist_b))/b->segments[j].length;
			bc->points[k].y = (b->points[j+1].y*dist_b+b->points[j].y*
				(b->segments[j].length-dist_b))/b->segments[j].length;
		} else if(pa>pb) {
			dist_a += pb*a->distance;
			dist_b = 0.f;
			ac->segments[k-1].length = pb*a->distance;
			bc->segments[k-1].length = db;
			bc->points[k] = b->points[++j];
			ac->points[k].x = (a->points[i+1].x*dist_a+a->points[i].x*
				(a->segments[i].length-dist_a))/a->segments[i].length;
			ac->points[k].y = (a->points[i+1].y*dist_a+a->points[i].y*
				(a->segments[i].length-dist_a))/a->segments[i].length;
		}

		/* Advance both points if they are equally spaced */
		else {
			dist_a = 0.f;
			dist_b = 0.f;
			ac->segments[k-1].length = da;
			bc->segments[k-1].length = db;
			ac->points[k] = a->points[++i];
			bc->points[k] = b->points[++j];
		}
	}
	ac->length = bc->length = k-1;
}

float osSymTrapezoid(osSymbol *a,osSymbol *b) {
	osSymbol *ac,*bc;
	float dist,dist_total=0.,dist_last,dx,dy;
	int i;
	ac = osSymNew();
	bc = osSymNew();
	createCompat(a,b,ac,bc);
	dx = ac->points[0].x-bc->points[0].x;
	dy = ac->points[0].y-bc->points[0].y;
	dist_last = sqrt(dx*dx+dy*dy);
	for(i=1;i<ac->length;i++) {
		dx = ac->points[i].x-bc->points[i].x;
		dy = ac->points[i].y-bc->points[i].y;
		dist = sqrt(dx*dx+dy*dy);
		dist_total += (dist_last+dist)*ac->segments[i-1].length/2;
		dist_last = dist;
	}
	osSymFree(ac);
	osSymFree(bc);
	return dist_total/(1.4142136*OS_SCALE*ac->distance);
}

float osSymTrapAngle(osSymbol *a,osSymbol *b) {
	osSymbol *ac,*bc;
	float dif_total=0.,dif_last,dif;
	int i;
	ac = osSymNew();
	bc = osSymNew();
	createCompat(a,b,ac,bc);
	dif_last = angleAbsDif(ac->segments[0].angle,bc->segments[0].angle);
	if(dif_last<0.f) dif_last = -dif_last;
	for(i=1;i<ac->length-1;i++) {
		dif = angleAbsDif(ac->segments[i].angle,bc->segments[i].angle);
		if(dif<0.f) dif = -dif;
		dif_total += (dif_last+dif)*ac->segments[i-1].length/2;
		dif_last = dif;
	}
	osSymFree(ac);
	osSymFree(bc);
	return dif_total/(2.f*M_PI*ac->distance);
}

float osSymTrapComp(osSymbol *a,osSymbol *b) {
	osSymbol *ac,*bc;
	float dist,dist_total=0.,dist_last,dx,dy,dif_total=0.,dif_last,dif;
	int i;
	ac = osSymNew();
	bc = osSymNew();
	createCompat(a,b,ac,bc);
	dx = ac->points[0].x-bc->points[0].x;
	dy = ac->points[0].y-bc->points[0].y;
	dist_last = sqrt(dx*dx+dy*dy);
	if(ac->length>0&&bc->length>0)
		dif_last = angleAbsDif(ac->segments[0].angle,bc->segments[0].angle);
	for(i=1;i<ac->length;i++) {

		/* Sum point distance */
		dx = ac->points[i].x-bc->points[i].x;
		dy = ac->points[i].y-bc->points[i].y;
		dist = sqrt(dx*dx+dy*dy);
		dist_total += (dist_last+dist)*ac->segments[i-1].length/2;
		dist_last = dist;

		/* Sum angle difference */
		if(i<ac->length-1) {
			dif = angleAbsDif(ac->segments[i].angle,bc->segments[i].angle);
			if(dif<0.f) dif = -dif;
			dif_total += (dif_last+dif)*ac->segments[i-1].length/2;
			dif_last = dif;
		}
	}
	osSymFree(ac);
	osSymFree(bc);
	return dist_total*dif_total/
		(2.8284271*OS_SCALE*M_PI*ac->distance*ac->distance);
}

void osSymTrain(osSymbol* symbol) {
	if(osSymValid(symbol)) {
		osSymProcess(os_cur_stroke,0,0);

		/* Free old arrays */
		if(symbol->points) free(symbol->points);
		if(symbol->segments) free(symbol->segments);

		/* Set variables */
		symbol->length = os_cur_stroke->length;
		symbol->distance = os_cur_stroke->distance;
		symbol->direction = os_cur_stroke->direction;

		/* Allocate memory */
		symbol->points = malloc(symbol->length*sizeof(osPoint));
		if(symbol->length>0)
			symbol->segments = malloc((symbol->length-1)*sizeof(osSegment));
		else symbol->segments = 0;

		/* Copy new arrays */
		memcpy(symbol->points,os_cur_stroke->points,
			symbol->length*sizeof(osPoint));
		if(symbol->segments) memcpy(symbol->segments,os_cur_stroke->segments,
			(symbol->length-1)*sizeof(osSegment));

		os_modified = 1;
	}
}

void osSymDump(osSymbol* sym) {
	if(osSymValid(sym)) {
		int i;
		char *text = osSymKeyString(sym);
		debug("Dumping '%s'...",text);
		for(i=0;i<sym->length;i++)
			debug("%d,%d",sym->points[i].x,sym->points[i].y);
	}
}

void osSymValidate(osSymbol *sym) {
	int i;
	float segdist=0.f;

	/* Sum segment distances */
	for(i=0;i<sym->length-1;i++) segdist += sym->segments[i].length;
	if(segdist!=sym->distance)
		debug("Segment distances do not sum correctly! %.1f sum, should be "
			"%.1f, difference of %.1f, difference per segment is %.2f",
			segdist,sym->distance,segdist-sym->distance,
			(segdist-sym->distance)/(sym->length-1));
}


/*
	Main
*/

int main(int argc,char *argv[]) {
	debug("Version " OS_VERSION " started (debug).");
	if(osSetCWD()) {
		os_cur_stroke = osSymNew();
		osFrontend(argc,argv);
		osProfileSave();
		debug("Program properly terminated.");
		return 0;
	}
	return 1;
}


/*
	Debugging
*/

#ifdef _DEBUG

void _debug(char* file,int line,char *format,...) {
	char buf[1024];
	va_list va;
	va_start(va,format);
	snprintf(buf,1023,"onestroke (%s:%d): %s\n",file,line,format);
	vfprintf(stderr,buf,va);
	va_end(va);
}

void _perrno(char* file,int line) {
	char buf[1024];
	snprintf(buf,1023,"onestroke (%s:%d)",file,line);
	perror(buf);
}

#endif
